/*
  
  File root.c - superblock and module routines for cdfs

  
  Copyright (c) 1999,2000 by Michiel Ronsse 

  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
  
*/

#include "cdfs.h"

#include <linux/module.h>

/********************************************************************/


static struct super_operations cdfs_ops;

int cdfs_parse_options(char *, cd *);

extern struct proc_dir_entry * proc_entry;

extern cd * proc_cd;

/***********************/

static struct super_block * cdfs_mount(struct super_block *sb, void *data, int silent){
  kdev_t dev = sb->s_dev;
  int i;
  extern struct file_operations * get_blkfops(unsigned int);
  struct cdrom_tochdr  hdr;
  struct cdrom_tocentry   entry;   
  int no_audio=0, no_data=0;
  cd * this_cd;

  PRINT("cdfs_mount\n");

  MOD_INC_USE_COUNT;

  lock_super(sb);

  set_blocksize(dev, SECTORSIZE);
  sb->s_blocksize = SECTORSIZE;
  sb->s_blocksize_bits = SECTORBITS;
  
  if (!(this_cd = sb->u.generic_sbp = kmalloc(sizeof(cd), GFP_KERNEL)))
      return NULL;

  this_cd->mode   = MODE; 
  this_cd->gid    = GID;
  this_cd->uid    = UID;
  this_cd->single = FALSE;

  this_cd->sector_cache = -READ_AHEAD-1;
  this_cd->audio_cache  = NULL;

  cdfs_parse_options((char *) data, this_cd);

  this_cd->ioctl = get_blkfops(MAJOR(dev))->ioctl;
  this_cd->inode_fake.i_rdev = this_cd->inode_fake.i_dev  = dev;
  
  if (cdfs_ioctl(sb, CDROMREADTOCHDR, (unsigned long)&hdr)){
    printk("ioctl(CDROMREADTOCHDR) failed\n");
    return NULL;
  }

  this_cd->nr_iso_sessions = 0;
  this_cd->size            = 0;
  this_cd->tracks          = hdr.cdth_trk1-hdr.cdth_trk0+1;

  PRINT("CD contains %d tracks: %d-%d\n", this_cd->tracks, hdr.cdth_trk0, hdr.cdth_trk1);

  /* reading track+ISO info */
  entry.cdte_format = CDROM_LBA;
  for (i=0; i<=this_cd->tracks; i++) {
    
    entry.cdte_track = (i==this_cd->tracks) ? CDROM_LEADOUT : i+1;
    PRINT("Read track %d\n", entry.cdte_track);
    
    if (cdfs_ioctl(sb, CDROMREADTOCENTRY, (unsigned long)&entry)){
      printk("ioctl(CDROMREADTOCENTRY) failed\n");
      return NULL;
    }

    this_cd->track[i].start_lba = entry.cdte_addr.lba;
    
    PRINT("Start[%d]: %d\n", i, this_cd->track[i].start_lba);

    if (i!=this_cd->tracks) {                 /* all tracks but the LEADOUT */
      if (entry.cdte_ctrl & CDROM_DATA_TRACK) {        
        no_data++;
        this_cd->track[i].iso_info  = get_iso_info(sb, i);
        this_cd->track[i].type      = DATA;
        if (this_cd->track[i].iso_info) {
          this_cd->track[i].time      = constructtime((char*)&(this_cd->track[i].iso_info->creation_date));
          this_cd->track[i].iso_size  = constructsize((char*)&(this_cd->track[i].iso_info->volume_space_size)) 
            * CD_FRAMESIZE;
          if (!this_cd->single) this_cd->track[i].iso_size += this_cd->track[i].start_lba * CD_FRAMESIZE;
          sprintf(this_cd->track[i].name, this_cd->single ? DATA_NAME_SINGLE : DATA_NAME_ISO, no_data);
          this_cd->lba_iso_sessions[this_cd->nr_iso_sessions].start = this_cd->track[i].start_lba;
          this_cd->lba_iso_sessions[this_cd->nr_iso_sessions].stop  = this_cd->track[i].iso_size/CD_FRAMESIZE;
          this_cd->nr_iso_sessions++;
        } else {  /* DATA, geen ISO */
          this_cd->track[i].time      = 0;
          this_cd->track[i].iso_size  = 0;		 
          sprintf(this_cd->track[i].name, DATA_NAME_UNK, no_data);
        }
      } else {
        no_audio++;
        this_cd->track[i].iso_info  = NULL;
        this_cd->track[i].type      = AUDIO;
        this_cd->track[i].time      = CURRENT_TIME;
        this_cd->track[i].iso_size  = 0;
        sprintf(this_cd->track[i].name, AUDIO_NAME, i+1 /*no_audio*/);
      }
    }
  }       
  
  PRINT("CD ends at %d\n", this_cd->track[this_cd->tracks].start_lba);

  
  /* Calculate track sizes */
  for (i=0; i<this_cd->tracks; i++){
    this_cd->track[i].stop_lba = this_cd->track[i+1].start_lba - 1;

    this_cd->track[i].track_size =  this_cd->track[i+1].start_lba -  this_cd->track[i].start_lba; /* in sectoren! */
      
    if (this_cd->track[i].type==DATA) {
      this_cd->track[i].track_size *= CD_FRAMESIZE;        
      this_cd->track[i].size = this_cd->track[i+1].start_lba * CD_FRAMESIZE;       
    } else {        
      this_cd->track[i].size = this_cd->track[i].track_size = 
        this_cd->track[i].track_size * CD_FRAMESIZE_RAW + WAV_HEADER_SIZE;
    }
    
    this_cd->size += this_cd->track[i].track_size;
    
    PRINT("Track %2d: (%dB)\n", i+1,  this_cd->track[i].size);      
  }  
  
  
  /* Check if CD is bootable*/
  if (this_cd->track[0].type==DATA) check_bootable(sb);

  /* Check for an HFS partition in the first data track */
  if (no_data) {
    i=0;
    while (i<this_cd->tracks) {
      if (this_cd->track[i].type==DATA)
        break;
      i++;
    }
    get_hfs_info(sb, i); 
  }
  
  PRINT("%d audio tracks and %d data tracks => %dbytes\n", 
        no_audio, no_data, this_cd->size);
  
  sb->s_magic  = CDFS_MAGIC;
  sb->s_flags |= MS_RDONLY;
  sb->s_op     = &cdfs_ops;
  sb->s_root   = d_alloc_root(iget(sb, 0), NULL);  /* was 1 ipv 0 */
  
  unlock_super(sb);

  proc_cd = this_cd; 

  return sb;
}

/************************************************************************/

static void cdfs_umount(struct super_block *sb) {
  int i;
  cd * this_cd = sb->u.generic_sbp;

  PRINT("cdfs_umount\n");

  for (i=0; i<=this_cd->tracks; i++)
    if ((this_cd->track[i].type == DATA) && this_cd->track[i].iso_info)
      kfree(this_cd->track[i].iso_info);

  if (this_cd->audio_cache) kfree(this_cd->audio_cache);
  this_cd->audio_cache=NULL;
  this_cd->sector_cache=-READ_AHEAD-1;

  kfree(sb->u.generic_sbp);

  MOD_DEC_USE_COUNT;
  proc_cd = NULL; 

  return;
}

/************************************************************************/

static int cdfs_statfs(struct super_block *sb, struct statfs *buf, int bufsize) {
  struct statfs tmp;
  cd * this_cd = sb->u.generic_sbp;
  PRINT("rmfs_statfs\n");

  memset(&tmp, 0, sizeof(tmp));
  tmp.f_type    = CDFS_MAGIC;
  tmp.f_bsize   = SECTORSIZE;
  tmp.f_blocks  = this_cd->size/SECTORSIZE;
  tmp.f_namelen = CDFS_MAXFN;
  tmp.f_files   = this_cd->tracks;
  return copy_to_user(buf, &tmp, bufsize) ? -EFAULT : 0;
}

/************************************************************************/

static int cdfs_readdir(struct file *filp, void *dirent, filldir_t filldir) {
  struct inode *inode = filp->f_dentry->d_inode;
  int i, pos=0;
  cd * this_cd = inode->i_sb->u.generic_sbp;

  PRINT("cdfs_readdir ino=%ld f_pos=%u\n", inode->i_ino, (int)filp->f_pos);
  
  if (!inode || !S_ISDIR(inode->i_mode)) return -ENOTDIR;
  
  if (filp->f_pos>=2+this_cd->tracks) return 1;

  switch ((unsigned long)filp->f_pos) {
  case 0:
    if (filldir(dirent, ".", 1, pos++, 1) <0) return 0;
    filp->f_pos++;
  case 1:
    if (filldir(dirent, "..", 2, pos++, 2) <0) return 0;
    filp->f_pos++;
  default:
    for(i=filp->f_pos; i<this_cd->tracks+2; i++) {
      if (filldir(dirent,
                  this_cd->track[i-2].name, 
                  strlen(this_cd->track[i-2].name),
                  pos, 
                  1+i) 
          <0) return 0;
      filp->f_pos++;
    }
  }
  return 1;
}

/************************************************************************/

static struct dentry * cdfs_lookup(struct inode *dir, struct dentry *dentry){
  unsigned long offset;
  int res;
  struct inode * inode;
  const char *name;
  int len;
  int i;
  unsigned ino=0;
  cd * this_cd = dir->i_sb->u.generic_sbp;

  PRINT("cdfs_lookup %s ino=%ld -> ", dentry->d_name.name, dir->i_ino);

  res = -ENOENT;
  if (!dir || !S_ISDIR(dir->i_mode))
    goto out;

  if (!(strcmp(".", dentry->d_name.name))) {
    PRINT("Found ."); 
    ino=1;
    res=0;
  } else if (!(strcmp("..", dentry->d_name.name))) {
    PRINT("Found ..");
    ino=2;
    res=0;
  } else {
    for(i=0; i<this_cd->tracks; i++)
      if (!(strcmp(this_cd->track[i].name, dentry->d_name.name))) {
        PRINT("Found %s", dentry->d_name.name);
        ino=3+i;
        res=0;
      }
  }

  if (res) goto out;

  offset = dir->i_ino; 
  name   = dentry->d_name.name;
  len    = dentry->d_name.len;

  /* use dentry cache */
  if ((inode = iget(dir->i_sb, ino))==NULL) {
    res = -EACCES;
  } else {
    d_add(dentry, inode); 
  }
  
  PRINT(" got inode %ld", inode->i_ino);

 out:
  PRINT("\n");
  return ERR_PTR(res);
}


/***************************************************************************/

static struct file_operations cdfs_dir_operations = {
  NULL,                                                     /* lseek - default */
  generic_file_read,	/* read */
  NULL,			/* write - bad */
  cdfs_readdir,		/* readdir */
  NULL,			/* poll - default */
  NULL,			/* ioctl */
  generic_file_mmap,	/* mmap */
  NULL,			/* open */
  NULL,			/* flush */
  NULL,			/* release */
  NULL,			/* fsync */
  NULL,			/* fasync */
  NULL,			/* check_media_change */
  NULL			/* revalidate */
};

static struct inode_operations cdfs_dir_inode_operations = {
  &cdfs_dir_operations,
  NULL,			/* create */
  cdfs_lookup,		/* lookup */
  NULL,			/* link */
  NULL,			/* unlink */
  NULL,			/* symlink */
  NULL,			/* mkdir */
  NULL,			/* rmdir */
  NULL,			/* mknod */
  NULL,			/* rename */
  NULL,			/* readlink */
  NULL,			/* follow_link */
  NULL,			/* readpage */
  NULL,			/* writepage */
  NULL,			/* bmap */
  NULL,			/* truncate */
  NULL,			/* permission */
  NULL,			/* smap */
};


/***********************************************************/

static void cdfs_read_inode(struct inode *i) {
  cd * this_cd = i->i_sb->u.generic_sbp;

  PRINT("read inode %ld\n", i->i_ino);

  if (i->i_ino <= 2) {                               /* . en .. */
    i->i_nlink = 2;                      /* Uuugh ?? */
    i->i_size  = 1234;                   /* Uuugh ?? */
    i->i_mtime = i->i_atime = i->i_ctime = CURRENT_TIME;
    i->i_uid   = this_cd->uid;
    i->i_gid   = this_cd->gid;
    i->i_mode  = S_IFDIR | S_IRUSR | S_IXUSR | S_IRGRP |  S_IXGRP | S_IROTH | S_IXOTH;
    i->i_op    = &cdfs_dir_inode_operations;
  } else {                                          /* file */
    i->i_nlink = 1;
    i->i_size  = this_cd->track[i->i_ino-3].size;
    i->i_mtime = i->i_atime = i->i_ctime = this_cd->track[i->i_ino-3].time;
    i->i_uid   = this_cd->uid;
    i->i_gid   = this_cd->gid;
    i->i_mode  = this_cd->mode;
    if ((this_cd->track[i->i_ino-3].type==DATA) && this_cd->track[i->i_ino-3].iso_size) {
      i->i_op    = &cdfs_file_inode_operations; 
    } else if (this_cd->track[i->i_ino-3].type==AUDIO) {
      i->i_op    = &cdda_file_inode_operations;
    } else if (this_cd->track[i->i_ino-3].type==BOOT) {
      i->i_op    = &cdfs_file_inode_operations;
    } else if (this_cd->track[i->i_ino-3].type==HFS) {
      if (this_cd->track[i->i_ino-3].hfs_offset)
        i->i_op    = &cdhfs_file_inode_operations; /* Bummer, this partition isn't properly aligned... */
      else
        i->i_op    = &cdfs_file_inode_operations;
    } else {
      i->i_op    = &cdXA_file_inode_operations;
    }
  }
}

/******************************************************************/

static struct super_operations cdfs_ops = {
  cdfs_read_inode,                      /* read inode */
  NULL,                                 /* write inode */
  NULL,                                 /* put inode */
  NULL,                                 /* delete inode */
  NULL,                                 /* notify change */
  cdfs_umount,                          /* put super */
  NULL,                                 /* write super */
  cdfs_statfs,                          /* statfs */
  NULL,                                 /* remount */
  NULL,                                 /* clear inode */
  NULL                                  /* umount begin */
};

static struct file_system_type cdfs_fs_type = {
  FSNAME,          /* const char *name; */
  FS_REQUIRES_DEV, /* int fs_flags;  */
  cdfs_mount,      /* struct super_block *(*read_super) (struct super_block *, void *, int); */
  NULL             /* struct file_system_type * next; */
};


/******************************************************/


MODULE_AUTHOR("Michiel Ronsse (ronsse@elis.rug.ac.be)");
MODULE_DESCRIPTION("CDfs: a CD filesystem");

EXPORT_NO_SYMBOLS;

int init_module(void) {
  PRINT("init_module (insmod)\n");
  if ((proc_entry = create_proc_entry(FSNAME, 0, NULL )))
    proc_entry->read_proc = read_proc;
  proc_cd=NULL;
  printk(FSNAME", "VERSION" loaded.\n");
  return register_filesystem(&cdfs_fs_type);
}

void cleanup_module(void) {
  PRINT("cleanup_module (rmmod)\n");
  if (proc_entry) remove_proc_entry(FSNAME , NULL);   
  unregister_filesystem(&cdfs_fs_type);
}



int cdfs_parse_options(char *options, cd * this_cd) {
  
  char *this_char,*value;
  
  /* from isofs */
  
  if (!options) return 1;
  
  for (this_char = strtok(options,","); this_char; this_char = strtok(NULL,",")) {
    
    if (!strcmp(this_char,"single")) 
      this_cd->single=TRUE;
    else {
      if ((value = strchr(this_char,'=')) != NULL)
        *value++ = 0;     
      if (value &&
          (!strcmp(this_char,"mode") ||
           !strcmp(this_char,"uid") ||
           !strcmp(this_char,"gid"))) {
        char * vpnt = value;                                   
        unsigned int ivalue = simple_strtoul(vpnt, &vpnt, 0);
        if (*vpnt) return 0;
        switch(*this_char) {
        case 'u':
          this_cd->uid = ivalue;
          break;
        case 'g':
          this_cd->gid = ivalue;
          break;
        case 'm':
          this_cd->mode = ivalue | S_IFREG ;
          break;
        }
      }
      else return 1;
    }
  }
  return 1;
}

                                  
